from hashlib import sha256
from pathlib import Path
from time import time
from typing import Optional

from plumbum import local

from ._service_api import MergeRequest, ServiceAPI

__all__ = ("DependencyUpdater", "NothingToUpdate", "MergeRequestExists",
           "AnotherMergeRequestExists", "UserNotFound",)


class _DependencyUpdaterException(Exception):
    pass


class NothingToUpdate(_DependencyUpdaterException):
    pass


class MergeRequestExists(_DependencyUpdaterException):
    pass


class AnotherMergeRequestExists(_DependencyUpdaterException):
    pass


class UserNotFound(_DependencyUpdaterException):
    pass


class DependencyUpdater:
    def __init__(self, service_api: ServiceAPI, pypi_index_url: str) -> None:
        self._service_api = service_api
        self._py_index_url = pypi_index_url

    def _get_file_hash(self, path: Path) -> str:
        with open(path, "r") as f:
            content = f.readlines()
        lines = [x for x in content if x.strip() and not x.strip().startswith("#")]
        return sha256("".join(lines).encode()).hexdigest()

    def _run_pip_compile(self, requirements_in: Path, requirements_out: Path) -> None:
        local["pip-compile"](requirements_in, "--upgrade",
                             "--index-url", self._py_index_url,
                             "--output-file", requirements_out)

    def update(self, requirements_in: Path, requirements_out: Path, *,
               assignee: Optional[str] = None,
               allow_multiple_mrs: bool = False,
               branch_prefix: str = "fresh-deps") -> MergeRequest:
        commit_message = "update dependencies"
        merge_request_title = "fresh-deps: update dependencies"

        assignee_id = None
        if assignee is not None:
            assignee_id = self._service_api.get_user_id(assignee)
            if assignee_id is None:
                raise UserNotFound(assignee)

        hash_before = self._get_file_hash(requirements_out)
        self._run_pip_compile(requirements_in, requirements_out)

        hash_after = self._get_file_hash(requirements_out)
        hash_short = hash_after[:10]
        if hash_after == hash_before:
            raise NothingToUpdate(hash_short)

        branch_name = "{}-{}-{}".format(branch_prefix, int(time()), hash_short)
        merge_requests = self._service_api.get_merge_requests()

        for merge_request in merge_requests:
            source_branch = merge_request.source_branch
            if source_branch.startswith(branch_prefix) and source_branch.endswith(hash_short):
                raise MergeRequestExists(merge_request.url)

        for merge_request in merge_requests:
            if merge_request.source_branch.startswith(branch_prefix) and (not allow_multiple_mrs):
                raise AnotherMergeRequestExists(merge_request.url)

        self._service_api.commit_file(requirements_out, commit_message, branch_name)
        mr = self._service_api.create_merge_request(branch_name, merge_request_title,
                                                    assignee_id=assignee_id)
        return mr



import argparse
from os import environ
from pathlib import Path
from typing import Any, Callable

from ._dependency_updater import (
    AnotherMergeRequestExists,
    DependencyUpdater,
    MergeRequestExists,
    NothingToUpdate,
)
from ._gitlab_api import GitLabAPI

__all__ = ("update_dependencies",)


def update_dependencies(logger: Callable[[str], Any] = print) -> None:
    parser = argparse.ArgumentParser("fresh-deps",
                                     description="Keep your Python dependencies fresh")

    parser.add_argument("requirements_in", type=Path,
                        help="Path to requirements.in")
    parser.add_argument("--output-file", type=Path, nargs="?", default=None,
                        help="Path to requirements.txt")

    default_pypi_index = "https://pypi.org/simple"
    parser.add_argument("--pypi-index-url", default=default_pypi_index,
                        help="PyPI index URL (default: {default_pypi_index})")

    default_gitlab_url = "https://gitlab.com"
    parser.add_argument("--gitlab-url",
                        default=environ.get("CI_SERVER_URL", default_gitlab_url),
                        help="GitLab server URL "
                             f"(default: $CI_SERVER_URL or '{default_gitlab_url}')")
    parser.add_argument("--gitlab-project-id",
                        default=environ.get("CI_PROJECT_ID", ""),
                        help="GitLab project ID (defaulT: $CI_PROJECT_ID)")
    default_gitlab_branch = "main"
    parser.add_argument("--gitlab-default-branch",
                        default=environ.get("CI_DEFAULT_BRANCH", default_gitlab_branch),
                        help="GitLab default branch "
                             f"(default: $CI_DEFAULT_BRANCH or '{default_gitlab_branch}')")
    docs_url = "https://docs.gitlab.com/ee/user/project/settings/project_access_tokens.html"
    parser.add_argument("--gitlab-private-token",
                        default=environ.get("CI_PRIVATE_TOKEN", ""),
                        help="GitLab private token "
                             f"(default: $CI_PRIVATE_TOKEN), documentation {docs_url}")
    parser.add_argument("--gitlab-assignee", help="GitLab assignee username (example: 'root')")
    parser.add_argument("--gitlab-allow-multiple-mrs", action="store_true", default=False,
                        help="Allow multiple opened merge requests")

    args = parser.parse_args()

    requirements_in = args.requirements_in.absolute().relative_to(Path.cwd())
    assert requirements_in.exists(), f"File '{requirements_in}' does not exist"

    if args.output_file is None:
        requirements_out = requirements_in.with_suffix(".txt")
    else:
        requirements_out = args.output_file.absolute().relative_to(Path.cwd())
    assert requirements_out.exists(), f"File '{requirements_out}' does not exist"

    assert args.gitlab_project_id, "Project ID is required"
    assert args.gitlab_private_token, "Private token is required"

    try:
        service_api = GitLabAPI(args.gitlab_url, args.gitlab_private_token,
                                args.gitlab_project_id, args.gitlab_default_branch)
    except BaseException as e:
        raise ConnectionError(f"Could not connect to GitLab: '{args.gitlab_url}'") from e

    dependency_updater = DependencyUpdater(service_api, args.pypi_index_url)
    try:
        merge_request = dependency_updater.update(requirements_in, requirements_out,
                                                  assignee=args.gitlab_assignee or None,
                                                  allow_multiple_mrs=args.gitlab_allow_multiple_mrs)
    except NothingToUpdate as e:
        logger(f"Nothing to update ({e})")
    except MergeRequestExists as e:
        logger(f"Merge request already exists ({e})")
    except AnotherMergeRequestExists as e:
        logger(f"Another merge request exists ({e})")
    else:
        logger(f"New merge request created: {merge_request.url}")



from pathlib import Path
from typing import List, Optional, Union

from gitlab import Gitlab

from ._service_api import MergeRequest, ServiceAPI

__all__ = ("GitLabAPI",)


class GitLabAPI(ServiceAPI):
    def __init__(self, server_url: str, private_token: str,
                 project_id: str, default_branch: str) -> None:
        self._gitlab = Gitlab(server_url, private_token, api_version="4")
        self._project_id = project_id
        self._default_branch = default_branch
        self._project = self._gitlab.projects.get(self._project_id)

    # Docs https://docs.gitlab.com/ee/api/users.html#list-users
    def get_user_id(self, username: str) -> Union[str, None]:
        users = self._gitlab.users.list(username=username)
        if len(users) == 0:
            return None
        return str(users[0].id)  # type: ignore

    # Docs https://docs.gitlab.com/ee/api/merge_requests.html#list-merge-requests
    def get_merge_requests(self) -> List[MergeRequest]:
        merge_requests = []
        for mr in self._project.mergerequests.list(state="opened", per_page=50, iterator=True):
            merge_requests.append(MergeRequest(mr.source_branch, mr.web_url))
        return merge_requests

    def _get_file_content(self, path: Path) -> str:
        with open(path) as f:
            return f.read()

    # https://docs.gitlab.com/ee/api/commits.html#create-a-commit-with-multiple-files-and-actions
    def commit_file(self, path: Path, message: str, source_branch: str) -> None:
        self._project.commits.create({
            "id": self._project_id,
            "branch": source_branch,
            "commit_message": message,
            "start_branch": self._default_branch,
            "actions": [
                {
                    "action": "update",
                    "file_path": str(path),
                    "content": self._get_file_content(path),
                }
            ]
        })

    # Docs https://docs.gitlab.com/ee/api/merge_requests.html#create-mr
    def create_merge_request(self, source_branch: str, title: str, *,
                             assignee_id: Optional[str] = None) -> MergeRequest:
        merge_request = self._project.mergerequests.create({
            "source_branch": source_branch,
            "target_branch": self._default_branch,
            "title": title,
            "remove_source_branch": True,
            "squash": True,
            "assignee_id": assignee_id,
        })
        return MergeRequest(source_branch, merge_request.web_url)



from ._dependency_updater import DependencyUpdater
from ._gitlab_api import GitLabAPI
from ._update_dependencies import update_dependencies

__all__ = ("update_dependencies", "GitLabAPI", "DependencyUpdater",)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Union

__all__ = ("ServiceAPI", "MergeRequest",)


@dataclass
class MergeRequest:
    source_branch: str
    url: str


class ServiceAPI(ABC):
    @abstractmethod
    def get_user_id(self, username: str) -> Union[str, None]:
        pass

    @abstractmethod
    def get_merge_requests(self) -> List[MergeRequest]:
        pass

    @abstractmethod
    def commit_file(self, path: Path, message: str, source_branch: str) -> None:
        pass

    @abstractmethod
    def create_merge_request(self, source_branch: str, title: str, *,
                             assignee_id: Optional[str] = None) -> MergeRequest:
        pass



from setuptools import find_packages, setup


def find_required():
   with open("requirements.txt") as f:
       return f.read().splitlines()

setup(
    name="freshdeps",
    version="1.0.0",
    description="Keep your Python dependencies fresh",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Nikita Tsvetkov",
    author_email="pychapter@2gis.ru",
    python_requires=">=3.8",
    url="https://github.com/Kater913/freshdeps",
    packages=find_packages(exclude=("tests",)),
    package_data={"fresh_deps": ["py.typed"]},
    entry_points={
        "console_scripts": [
            "fresh-deps = fresh_deps:update_dependencies",
            "fresh_deps = fresh_deps:update_dependencies",
        ],
    },
    install_requires=find_required(),
)
